@@ -36,14 +36,34 @@
 :-module(saml,
          [saml_authenticate/4]).
 
-:- use_module(library(sgml)).
-:- use_module(library(base64)).
-:- use_module(library(zlib)).
-:- use_module(library(xmldsig)).
-:- use_module(library(xmlenc)).
-:- use_module(library(http/http_path)).
-:- use_module(library(http/http_dispatch)).
-:- use_module(library(http/http_client)).
+:- autoload(library(base64),[base64/2]).
+:- autoload(library(crypto),[rsa_sign/4,hex_bytes/2]).
+:- autoload(library(debug),[debug/3,debugging/1]).
+:- autoload(library(error),
+	    [domain_error/2,existence_error/2,permission_error/3]).
+:- autoload(library(lists),[member/2,subtract/3,select/3]).
+:- autoload(library(memfile),
+	    [ new_memory_file/1,
+	      open_memory_file/4,
+	      memory_file_to_atom/2,
+	      free_memory_file/1
+	    ]).
+:- autoload(library(quintus),[otherwise/0]).
+:- autoload(library(sgml),[load_structure/3]).
+:- autoload(library(sgml_write),[xml_write/3]).
+:- autoload(library(sha),[sha_hash/3]).
+:- autoload(library(ssl),
+	    [load_private_key/3,load_certificate/2,same_certificate/2]).
+:- autoload(library(url),[parse_url/2,parse_url_search/2]).
+:- autoload(library(uuid),[uuid/1]).
+:- autoload(library(xmldsig),[xmld_verify_signature/4]).
+:- autoload(library(xmlenc),
+	    [load_certificate_from_base64_string/2,decrypt_xml/4]).
+:- autoload(library(zlib),[zopen/3]).
+:- autoload(library(http/http_client),[http_read_data/3]).
+:- autoload(library(http/http_dispatch),[http_redirect/3]).
+:- autoload(library(http/http_path),[http_absolute_location/3]).
+:- autoload(library(http/http_open),[http_open/3]).
 
 /** <module> SAML Authentication
 
@@ -280,18 +300,20 @@ saml_simple_sign(PrivateKey, XMLString, _SAMLRequest, RelayState, ['SigAlg'=SigA
     base64(RawSignature, Signature),
     debug(saml, 'Signature:~n~w~n', [Signature]).
 
-saml_sign(PrivateKey, _XMLString, SAMLRequest, RelayState, ['SigAlg'=SigAlg,'Signature'=Signature]):-
+saml_sign(PrivateKey, _XMLString, SAMLRequest, RelayState, ['SigAlg'=SigAlg,'Signature'=Base64Signature]):-
     SigAlg = 'http://www.w3.org/2000/09/xmldsig#rsa-sha1',
     parse_url_search(CodesToSign, ['SAMLRequest'=SAMLRequest, 'RelayState'=RelayState, 'SigAlg'=SigAlg]),
     string_codes(DataToSign, CodesToSign),
     debug(saml, 'Data to sign with HTTP-Redirect binding:~n~s~n', [DataToSign]),
     sha_hash(DataToSign, Digest, [algorithm(sha1)]),
-    rsa_sign(PrivateKey, Digest, RawSignature,
+    rsa_sign(PrivateKey, Digest, HexSignature,
              [ type(sha1),
                encoding(octet)
              ]),
-    base64(RawSignature, Signature),
-    debug(saml, '~nSignature:~n~w~n', [Signature]).
+    hex_bytes(HexSignature, SignatureBytes),
+    atom_codes(SignatureAtom, SignatureBytes),
+    base64(SignatureAtom, Base64Signature),
+    debug(saml, '~nSignature:~n~w~n', [Base64Signature]).
 
 saml_acs_handler(ServiceProvider, Options, Request):-
     debug(saml, 'Got a message back from IdP!~n', []),
@@ -373,7 +395,8 @@ process_saml_response(XML0, ServiceProvider, Callback, RequestURL, Options):-
     ( member(element(ns(_, DS):'Signature', _, Signature), Response)->
         xmld_verify_signature(XML, Signature, Certificate, []),
         % Check that the certificate used to sign was one in the metadata
-        (  saml_idp_certificate(ServiceProvider, IssuerName, signing, Certificate)
+        (  saml_idp_certificate(ServiceProvider, IssuerName, signing, IDPCertificate),
+           same_certificate(Certificate, IDPCertificate)
         -> true
         ;  domain_error(trusted_certificate, Certificate)
         )
@@ -427,7 +450,8 @@ process_assertion(ServiceProvider, _EntityID, Document, Attributes, Assertion, A
     ( member(element(DS:'Signature', _, Signature), Assertion)->
         xmld_verify_signature(Document, Signature, Certificate, []),
         % Check that the certificate used to sign was one in the metadata
-        (  saml_idp_certificate(ServiceProvider, IssuerName, signing, Certificate)
+        (  saml_idp_certificate(ServiceProvider, IssuerName, signing, IDPCertificate),
+           same_certificate(Certificate, IDPCertificate)
         -> true
         ;  domain_error(trusted_certificate, Certificate)
         )
